/*
 * Freescale PWM Register Definitions
 *
 * Copyright 2011 Freescale Semiconductor, Inc.
 *
 * Author: Chunhe Lan <Chunhe.Lan@freescale.com>
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __ARCH_FSL_PWM_H
#define __ARCH_FSL_PWM_H

#define FSL_PWMCR_STOPEN			(1 << 25)
#define FSL_PWMCR_DOZEEN			(1 << 24)
#define FSL_PWMCR_WAITEN			(1 << 23)
#define FSL_PWMCR_DEBUGEN		        (1 << 22)
#define FSL_PWMCR_BCTR				(1 << 21)
#define FSL_PWMCR_HCTR				(1 << 20)
#define FSL_PWMCR_POUTC_HIGHT			(0 << 18)
#define FSL_PWMCR_POUTC_LOW			(1 << 18)
#define FSL_PWMCR_CLKSRC        	 	(1 << 16)
#define FSL_PWMCR_PRESCALER(x)			(((x - 1) & 0xFFF) << 4)
#define FSL_MAX_PRESCALER			0x00000FFF
#define FSL_PWMCR_SWR				(1 << 3)
#define FSL_PWMCR_REPEAT_ONE			(0 << 1)
#define FSL_PWMCR_REPEAT_TWO			(1 << 1)
#define FSL_PWMCR_REPEAT_FOUR			(2 << 1)
#define FSL_PWMCR_REPEAT_EIGHT			(3 << 1)
#define FSL_PWMCR_EN				(1 << 0)

#define FSL_PWMSR_ALL_MASK			0x0000007F
#define FSL_PWMSR_FWE_CMP_ROV_MASK		0x00000070
#define FSL_PWMSR_FWE				(1 << 6)
#define FSL_PWMSR_CMP				(1 << 5)
#define FSL_PWMSR_ROV				(1 << 4)
#define FSL_PWMSR_FE				(1 << 3)
#define FSL_PWMSR_FIFOAV			(7 << 0)

#define FSL_PWMIR				(7 << 0)
#define FSL_PWMIR_CIE				(1 << 2)
#define FSL_PWMIR_RIE				(1 << 1)
#define FSL_PWMIR_FIE				(1 << 0)

#define FSL_PMUXCR1_SPI1_ANT_TCXO_PWM_GPIO	(3 << 0)
#define FSL_PMUXCR1_ANT_TCXO_PWM_GPIO		(1 << 0)
#define FSL_PMUXCR2_UART_PWM_DSP_GPIO		(3 << 28)
#define FSL_PMUXCR2_PWM_GPIO			(1 << 28)

#define FSL_DEVDISR2_PWM1			(1 << 23)
#define FSL_DEVDISR2_PWM1_EN			(0 << 23)
#define FSL_DEVDISR2_PWM2			(1 << 22)
#define FSL_DEVDISR2_PWM2_EN			(0 << 22)

struct pwm_reg {
	u32 pwmcr;     	    /* PWM Control Register */
	u32 pwmsr;     	    /* PWM Status Register */
	u32 pwmir;     	    /* PWM Interrupt Register */
	u32 pwmsar;         /* PWM Sample Register */
	u32 pwmpr;          /* PWM Period Register */
	u32 pwmcnr;         /* PWM Counter Register */
};

struct pwm_device {
	struct list_head node;
	struct device dev;

	const char *label;
	struct clk *clk;
	int clk_enabled;
	struct pwm_reg __iomem *regs;
	int irq;

	unsigned int use_count;
	unsigned int pwm_id;
	int duty;
	int period;
	int pwmo_invert;
	void (*enable_pwm_pad)(void);
	void (*disable_pwm_pad)(void);
};

struct gubr {
	u32 res24[0x18];
	u32 pmuxcr1;
	u32 pmuxcr2;
	u32 res3[0x03];
	u32 devdisr2;
};

extern int pwm_config(struct pwm_device *pwm, int duty_ns, int period_ns);
extern int pwm_enable(struct pwm_device *pwm);
extern void pwm_disable(struct pwm_device *pwm);
extern struct pwm_device *pwm_request(int pwm_id, const char *label);
extern void pwm_free(struct pwm_device *pwm);

#endif /* __ARCH_FSL_PWM_H */
