/*
 * BSC9131RDB, BSC9131QDS, BSC9132 Board Setup
 *
 * Author: Priyanka Jain <Priyanka.Jain@freescale.com>
 *
 * Copyright 2011-2012 Freescale Semiconductor, Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

#include <linux/of_platform.h>
#include <linux/lmb.h>
#include <linux/pci.h>

#include <asm/mpic.h>
#include <asm/swiotlb.h>
#include <asm/pci-bridge.h>

#include <sysdev/fsl_soc.h>
#include <sysdev/fsl_pci.h>

void __init bsc913x_board_pic_init(void)
{
	struct mpic *mpic;
	struct resource r;
	struct device_node *np;

	np = of_find_node_by_type(NULL, "open-pic");
	if (np == NULL) {
		printk(KERN_ERR "Could not find open-pic node\n");
		return;
	}

	if (of_address_to_resource(np, 0, &r)) {
		printk(KERN_ERR "Failed to map mpic register space\n");
		of_node_put(np);
		return;
	}

	mpic = mpic_alloc(np, r.start, MPIC_PRIMARY | MPIC_WANTS_RESET |
	  MPIC_BIG_ENDIAN | MPIC_BROKEN_FRR_NIRQS | MPIC_SINGLE_DEST_CPU,
	  0, 256, " OpenPIC  ");

	BUG_ON(mpic == NULL);
	of_node_put(np);

	mpic_init(mpic);

}

/*
 * Setup the architecture
 */

static void __init bsc913x_board_setup_arch(void)
{
#ifdef CONFIG_PCI
	struct device_node *np;
	struct pci_controller *hose;
#endif
	dma_addr_t max = 0xffffffff;

	if (ppc_md.progress)
		ppc_md.progress("bsc913x_board_setup_arch()", 0);

#ifdef CONFIG_PCI
	for_each_node_by_type(np, "pci") {
		if (of_device_is_compatible(np, "fsl,bsc9132-pcie")) {
			struct resource rsrc;
			of_address_to_resource(np, 0, &rsrc);

			fsl_add_bridge(np, 0);
			hose = pci_find_hose_for_OF_device(np);
			max = min(max, hose->dma_window_base_cur +
					hose->dma_window_size);
		}
	}
#endif

#ifdef CONFIG_SWIOTLB
	if (lmb_end_of_DRAM() > max) {
		ppc_swiotlb_enable = 1;
		set_pci_dma_ops(&swiotlb_dma_ops);
		ppc_md.pci_dma_dev_setup = pci_dma_dev_setup_swiotlb;
	}
#endif

#if defined(CONFIG_SMP)
	mpc85xx_smp_init();
#endif

	printk(KERN_INFO "bsc913x board from Freescale Semiconductor\n");
}

static struct of_device_id __initdata bsc913x_board_ids[] = {
	{ .type = "soc", },
	{ .type = "dsp", },
	{ .compatible = "soc", },
	{ .compatible = "simple-bus", },
	{ .compatible = "gianfar", },
	{},
};

static int __init bsc913x_board_publish_devices(void)
{
	return of_platform_bus_probe(NULL, bsc913x_board_ids, NULL);
}
machine_device_initcall(bsc9131_qds, bsc913x_board_publish_devices);
machine_device_initcall(bsc9131_rdb, bsc913x_board_publish_devices);
machine_device_initcall(bsc9132_qds, bsc913x_board_publish_devices);

/*
 * Called very early, device-tree isn't unflattened
 */
static int __init bsc9131_qds_probe(void)
{
	unsigned long root = of_get_flat_dt_root();

	if (of_flat_dt_is_compatible(root, "fsl,bsc9131qds"))
		return 1;
	return 0;
}

static int __init bsc9131_rdb_probe(void)
{
	unsigned long root = of_get_flat_dt_root();

	if (of_flat_dt_is_compatible(root, "fsl,bsc9131rdb"))
		return 1;
	return 0;
}

static int __init bsc9132_qds_probe(void)
{
	unsigned long root = of_get_flat_dt_root();

	if (of_flat_dt_is_compatible(root, "fsl,bsc9132qds"))
		return 1;
	return 0;
}

define_machine(bsc9131_qds) {
	.name			= "BSC9131 QDS",
	.probe			= bsc9131_qds_probe,
	.setup_arch		= bsc913x_board_setup_arch,
	.init_IRQ		= bsc913x_board_pic_init,
	.get_irq		= mpic_get_irq,
	.restart		= fsl_rstcr_restart,
	.calibrate_decr		= generic_calibrate_decr,
	.progress		= udbg_progress,
};

define_machine(bsc9131_rdb) {
	.name			= "BSC9131 RDB",
	.probe			= bsc9131_rdb_probe,
	.setup_arch		= bsc913x_board_setup_arch,
	.init_IRQ		= bsc913x_board_pic_init,
	.get_irq		= mpic_get_irq,
	.restart		= fsl_rstcr_restart,
	.calibrate_decr		= generic_calibrate_decr,
	.progress		= udbg_progress,
};

define_machine(bsc9132_qds) {
	.name			= "BSC9132 QDS",
	.probe			= bsc9132_qds_probe,
	.setup_arch		= bsc913x_board_setup_arch,
	.init_IRQ		= bsc913x_board_pic_init,
#ifdef CONFIG_PCI
	.pcibios_fixup_bus	= fsl_pcibios_fixup_bus,
#endif
	.get_irq		= mpic_get_irq,
	.restart		= fsl_rstcr_restart,
	.calibrate_decr		= generic_calibrate_decr,
	.progress		= udbg_progress,
};
