/*
 * timesync framework
 *
 * Author: bhaskar upadhaya <bhaskar.upadhaya@freescale.com>
 *
 * Copyright 2011 Freescale Semiconductor, Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

#ifndef __TIME_SYNC_H__
#define __TIME_SYNC_H__

typedef unsigned long long UInteger64;
typedef unsigned int UInteger32;

#define FIRST_GPS_PULSE		0x01

#define TMR_TEVENT_ETS2		0x02000000
#define TMR_TEVENT_ETS1		0x01000000
#define TMR_TEVENT_ALM2		0x00020000
#define TMR_TEVENT_ALM1		0x00010000
#define TMR_TEVENT_PP1		0x00000080
#define TMR_TEVENT_PP2		0x00000040
#define TMR_TEVENT_PP3		0x00000020

#define TMR_TEMASK_ETS2		0x02000000
#define TMR_TEMASK_ETS1		0x01000000
#define TMR_TEMASK_ALM2		0x00020000
#define TMR_TEMASK_ALM1		0x00010000
#define TMR_TEMASK_PP1		0x00000080
#define TMR_TEMASK_PP2		0x00000040

#define TMR_STAT_ETS1_VLD	0x01000000

#define TMR_CTRL_BYPASS_ADDEND	0x00000008

#define EXTERNAL_CLK_FREQ 19200000
#define MAX_32_REG_LIMIT 0xffffffff
#define OVERFLOW_LIMIT 223
#define FIPER_START_DELAY 	1 /* seconds */

#define NSEC	((UInteger64)1000000000)
#define	BOARD_TIME_TO_SEC(board_time)	((UInteger32)(board_time/NSEC))

#define BOARD_TIME_TO_NSEC(board_time, board_time_sec)\
	((UInteger32)(board_time - (((UInteger64)board_time_sec) * NSEC)))

#define BUFFER_SIZE 200

#define ETSEC_NODE "fsl,etsec2"
#define DAC_CORRECTION "dac"
#define ADDEND_CORRECTION "addend"
#define DAC_ADDEND_CORRECTION "dac_addend"

#define RAW_GPS "raw_gps"
#define GPS_PPS "gps_pps"
#define PTP_PPS "ptp_pps"
#define NLM_PPS "nlm_pps"

#define VERSION_NR 1

enum {
	TRUE = 0,
	FALSE
};

enum time_sync_events {
	BLACK_OUT_DURATION,
	TIME_SYNC_SOURCE,
	LTE_DELAY,
	LAST_BLACKOUT_OCCURENCE,
	NEW_MODE_AFTER_BLACKOUT,
	STATS,
	CORRECTION_MODE
};

enum time_sync_src {
	PPS_SRC_GPS = 1,
	PPS_SRC_RAW_GPS,
	PPS_SRC_PTP
};

struct time_sync_stats {
	u32 gps_cnt;
	u32 fiper1_cnt;
	u32 alarm1_cnt;
	u32 alarm2_cnt;
};
struct time_sync_priv {
	/* two states are possible
	 *	1. GPS_PPS_INPROGRESS
	 *	2. GPS_PPS_BLACKOUT
	 */
	u32 state;
	u32 old_jiffy;
	bool awaiting_first_gps;
	u8 first_blackout;
	wait_queue_head_t wait_h;
	struct kobject *time_sync_kobj;
	unsigned int irq;
	struct time_sync_data *time_sync;
	struct time_sync_stats stats;
	struct timespec blackout_time;
};

struct time_sync_sysfs {
	struct attribute attr;
	char data[BUFFER_SIZE];
	u32 cmd_type;
};

struct time_sync_lock {
	struct mutex lock;
	u8 condition; /* ISR sets it to 1 to indicate rx of timing pulse */
	u32 g_gps_cnt;
};

int do_time_sync_ioctl(struct net_device *dev, struct ifreq *ifr);
#endif
