/*
 * drivers/rf/phy/ad9361.h
 *
 * Copyright 2011-2012 Freescale Semiconductor, Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

#ifndef AD_PHY_H
#define AD_PHY_H

#define MAX_CARRIER_FREQ_KHZ	6000000
#define MIN_CARRIER_FREQ_KHZ	47000

/* For creating instruction word/command */
#define OPCODE_WRITE 0x8000
#define OPCODE_READ 0x0000
#define BYTES_TRANSFER_MASK 0x7000
#define SHIFT_BYTES_TRANSFER 12
#define REG_ADDRESS_MASK 0x03FF
#define COMMAND_MASK 0xFFFF0000
#define COMMAND_LEN 2

#define TRANSACTION_BYTES 3
#define MAX_READ_TRANSACTION_SIZE 	8
#define RXBUF_SIZE			10

#define BBPLL_LOCK_MASK 0x80
#define PRODUCT_CODE_REG 0x37
#define PRODUCT_ID_MASK 0x08
#define PRODUCT_ID_9361 0x08
#define REV_MASK	0x07

/* Calibration status Registers */
#define REG_CH1_OVERFLOW	0x05E
#define REG_RX_CAL_STATUS	0x244
#define REG_TX_CAL_STATUS	0x284
#define REG_CALIBRATION_CONTROL	0x016
#define REG_CALIBRATION_CONFIG1	0x169
#define REG_RX_CP_CONFIG	0x23D
#define REG_TX_CP_CONFIG	0x27D

#define VAL_CAL_CONF1_TRACKOFF	0xC0

/* Calibration status masks */
#define MASK_BBPLL_LOCK		0x80
#define MASK_RX_CP_CAL_VALID	0x80
#define MASK_TX_CP_CAL_VALID	0x80
#define MASK_RX_BB_TUNE		0x80
#define MASK_TX_BB_TUNE		0x40
#define MASK_DC_CAL_BBSTART	0x00
#define MASK_DC_CAL_RFSTART	0x01
#define MASK_TXQUAD_CAL		0x10

/*
 * 24 = 0x18 which is the AuxDAC1 word address
 * 26 = 0x1A which is the AuxDAC1 config address
 */
#define DAC1_WORD 	24
#define DAC1_CONFIG  	26

/* TX Attenuation Registers */
#define TX1_ATTEN0	0x073
#define TX1_ATTEN1	0x074
#define TX2_ATTEN0	0x075
#define TX2_ATTEN1	0x076
#define ATTEN0_MASK	0x000000FF
#define ATTEN1_MASK	0x00000100
#define ATTEN_MSB_BIT_MASK	0x01
#define MSB_SHIFT	8

#define TX_ENABLE	0x01
#define TX_DISABLE	0x00
#define TX_ENABLE_REG	0x002
#define TX1_ENABLE_MASK	0x40
#define TX2_ENABLE_MASK	0x80

#define RX_ENABLE	0x01
#define RX_DISABLE	0x00
#define RX_ENABLE_REG	0x003
#define RX1_ENABLE_MASK	0x40
#define RX2_ENABLE_MASK	0x80

#define RSSI_READBACK_REG 0x1A7
#define RSSI_CONFIG_REG	0x158
#define RSSI_DELAY_REG	0x156
#define RSSI_MEAS_DUR_10_REG	0x150
#define RSSI_MEAS_DUR_32_REG	0x151
#define RSSI_WEIGHT0_REG	0x152
#define RSSI_WEIGHT1_REG	0x153
#define RSSI_WEIGHT2_REG	0x154
#define RSSI_WEIGHT3_REG	0x155

#define RSSI_MAX_WEIGHT		255
#define RSSI_MEAS_MODE_MASK	0xE2
#define RSSI_GAIN_CHANGE_EN_AGC_MODE	0x14
/* RSSI delay reg value is decremented by RX sample rate divided by 8*/
#define RSSI_DELAY_5MHZ		(256 / 8)
#define RSSI_DELAY_10MHZ	(512 / 8)
#define RSSI_DELAY_15MHZ	(256 / 8)
#define RSSI_DELAY_20MHZ	(1024 / 8)

#define SUBFRAME_SIZE_5MHZ	7680
#define SUBFRAME_SIZE_10MHZ	15360
#define SUBFRAME_SIZE_15MHZ	23040
#define SUBFRAME_SIZE_20MHZ	30720

/* For 9 bit RSSI symbol/preamble value RSSI is equivalent to 0.25dB/LSB.
 * Since we can not do floating pt operations in kernel, we multiply
 * Resolution with RSSI_MULTIPLER, and pass this multipler to user space
 * which can convert rssi to floating pt again.
 */
#define RSSI_MULTIPLIER	100
#define RSSI_RESOLUTION	((int) (0.25 * RSSI_MULTIPLIER))
#define LSB_SHIFT	1
#define RSSI_LSB_MASK1	0x01
#define RSSI_LSB_MASK2	0x02

#define REG_RXEN_N_FILTER_CTRL		0x003
#define RX1_EN				0x40
#define RX2_EN				0x80

/* Gain index read/write register for MGC */
#define REG_RX1_MGC_FULL_TBL_IDX	0x109
#define REG_RX2_MGC_FULL_TBL_IDX	0x10c

/* Gain Index read back registers. They are
 * applicable for both AGC and MGC
 */
#define REG_RX1_FULL_TBL_IDX		0x2B0
#define REG_RX2_FULL_TBL_IDX		0x2B5
#define FULL_TBL_IDX_MASK		0x7f

/*Fast attack state register*/
#define REG_FAST_ATK_STATE		0x2B3
#define FAST_ATK_MASK			0x7
#define RX1_FAST_ATK_SHIFT		0
#define RX2_FAST_ATK_SHIFT		4

/*Fast attack state machine states*/
#define FAST_ATK_RESET			0
#define FAST_ATK_PEAK_DETECT		1
#define FAST_ATK_PWR_MEASURE		2
#define FAST_ATK_FINAL_SETTELING	3
#define FAST_ATK_FINAL_OVER		4
#define FAST_ATK_GAIN_LOCKED		5

#define REG_AGC_CONF1			0x0FA

#define RX_GAIN_CTL_MASK		0x03
#define RX2_GAIN_CTL_SHIFT		2
#define RX1_GAIN_CTL_SHIFT		0

#define RX_GAIN_CTL_MGC				0x00
#define RX_GAIN_CTL_AGC_FAST_ATK		0x01
#define RX_GAIN_CTL_AGC_SLOW_ATK		0x02
#define RX_GAIN_CTL_AGC_SLOW_ATK_HYBD		0x03
#define SLOW_ATK_HYBD_BIT_EN			0x10

#define REG_AGC_CONF2			0x0FB
#define FULL_GAIN_TBL			0x08

#define RXGAIN_FULL_TBL_MAX_IDX		90
#define RXGAIN_SPLIT_TBL_MAX_IDX	40

/*ENSM config1 register*/
#define REG_ENSM_CONF1			0x014
#define ENSM_CONF1_TO_ALERT		(1 << 0)
#define ENSM_CONF1_AUTO_GAIN_LOCK	(1 << 1)
#define ENSM_CONF1_FORCE_ALERT		(1 << 2)
#define ENSM_CONF1_LEVEL_MODE		(1 << 3)
#define ENSM_CONF1_ENSM_PIN_CTL_EN	(1 << 4)
#define ENSM_CONF1_FORCE_TX_ON		(1 << 5)
#define ENSM_CONF1_FORCE_RX_ON		(1 << 6)
#define ENSM_CONF_RX_EN_CAL		(1 << 7)

/*ENSM state - Read only*/
#define REG_DEV_STATE			0x017
#define ENSM_STATE_SHIFT		0x0
#define ENSM_STATE_MASK			0x0f

#define ENSM_STATE_SLEEP_WAIT		0x0
#define ENSM_STATE_ALERT		0x5
#define ENSM_STATE_TX			0x6
#define ENSM_STATE_TX_FLUSH		0x7
#define ENSM_STATE_RX			0x8
#define ENSM_STATE_RX_FLUSH		0x9
#define ENSM_STATE_FDD			0xA
#define ENSM_STATE_FDD_FLUSH		0xB
#define ENSM_STATE_INVALID		0xff

/* Synthesizer Registers */
#define REG_RFPLL_DIVIDER		0x005
#define REG_VCO_OUTPUT			0x23A
#define REG_VCO_VARACTOR		0x239
#define REG_VCO_BIAS1			0x242
#define REG_FORCE_VCO_TUNE1		0x238
#define REG_VCO_VARACTOR_CONTROL1	0x251
#define REG_CP_CURRENT			0x23B
#define REG_LOOP_FILTER1		0x23E
#define REG_LOOP_FILTER2		0x23F
#define REG_LOOP_FILTER3		0x240
#define REG_INTEGER_BYTE0		0X231
#define REG_INTEGER_BYTE1		0X232
#define REG_FRACTIONAL_BYTE0		0X233
#define REG_FRACTIONAL_BYTE1		0X234
#define REG_FRACTIONAL_BYTE2		0X235
#define REG_REF_DIVIDE_CONFIG1		0x2AB
#define REG_REF_DIVIDE_CONFIG2		0x2AC

#define RX_REG_OFFSET			0x000
#define TX_REG_OFFSET			0x040

#define RX_SCALER_MASK1			0x01
#define RX_SCALER_MASK2			0x80
#define TX_SCALER_MASK			0x0C

#define DEVICE_REF_FREQ_KHZ		19200
#define RFPLL_MODULUS			8388593

#define MASK_VCO_OUTPUT			0xF0
#define MASK_VCO_VARACTOR		0xF0
#define MASK_VCO_BIAS1			0xE0
#define MASK_FORCE_VCO_TUNE1		0x87
#define MASK_VCO_VARACTOR_CONTROL1	0xF0
#define MASK_CP_CURRENT			0xC0
#define MASK_LOOP_FILTER3		0xF0
#define MASK_FRAC_B0			0xff
#define MASK_FRAC_B1			0xff00
#define MASK_FRAC_B2			0xff0000
#define MASK_INT_B0			0xff
#define MASK_INTEGER_REG1		0xf8
#define MASK_INT_B1			0x700
#define SHIFT_FRAC_B0			0
#define SHIFT_FRAC_B1			8
#define SHIFT_FRAC_B2			16
#define SHIFT_INT_B1			8
#define SHIFT_VCO_CAL_OFFSET		3
#define SHIFT_VCO_BIAS_TCF		3
#define SHIFT_LOOP_FILTER_C2		4
#define SHIFT_LOOP_FILTER_R1		4

enum rx_gain_table_type {
	RXGAIN_FULL_TBL,
	RXGAIN_SPLIT_TBL,
};

enum rx_gain_table_name {
	FULL_TBL_200_1300_MHZ,
	FULL_TBL_1300_4000_MHZ,
	FULL_TBL_4000_6000_MHZ,
	RXGAIN_TBLS_END,
};

enum rf_synth_regs {
	VCO_Output_Level,
	VCO_Varactor,
	VCO_Bias_Ref,
	VCO_Bias_Tcf,
	VCO_Cal_Offset,
	VCO_Varactor_Reference,
	Charge_Pump_Current,
	Loop_Filter_C2,
	Loop_Filter_C1,
	Loop_Filter_R1,
	Loop_Filter_C3,
	Loop_Filter_R3
};

enum rf_synth_params {
	Band,
	Reference_Frequency,
	Loop_Bandwidth,
	Index,
	VCO_Frequency,
	VCO_Kv
};

struct rx_gain_info {
	enum rx_gain_table_type tbl_type;
	int starting_gain_db;
	int max_gain_db;
	int gain_step_db;
	int max_idx;
	int idx_step_offset;
};

struct ad_dev_info {
	struct spi_device *ad_spi;
	u8 rx_buf[10];
	u8 prev_ensm_state;
	int ensm_pin_ctl_en;
	unsigned int freq_band;
	enum rf_network_mode net_mode;
	struct rx_gain_info rx_gain[RXGAIN_TBLS_END];
	struct	rf_synth_table synth_table;
	int reset_gpio;
	int pa_en_gpio;
	int lna_en_gpio;
};

#endif /* AD_PHY_H */
