/*
 * PACEKT_UM	An implementation of the PACKET_UM socket type. A derivative
 *		of PACKET_MMAP socket.
 *
 * Authors:	Himanshu Seth <b21215@freescale.com>
 *		Hemant Agrawal <hemant@freescale.com>
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 *	Copyright 2011-2012 Freescale Semiconductor, Inc.
*/

#include <linux/types.h>
#include <linux/mm.h>
#include <linux/capability.h>
#include <linux/fcntl.h>
#include <linux/socket.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <linux/netdevice.h>
#include <linux/if_packet.h>
#include <linux/wireless.h>
#include <linux/kernel.h>
#include <linux/kmod.h>
#include <net/net_namespace.h>
#include <net/ip.h>
#include <net/protocol.h>
#include <linux/skbuff.h>
#include <net/sock.h>
#include <linux/errno.h>
#include <linux/timer.h>
#include <asm/system.h>
#include <linux/uaccess.h>
#include <asm/ioctls.h>
#include <asm/page.h>
#include <asm/cacheflush.h>
#include <linux/io.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <linux/poll.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/mutex.h>
#include <linux/if_vlan.h>
#include <linux/if_pmal.h>
#include "af_packet.h"

#ifdef CONFIG_INET
#include <net/inet_common.h>
#endif

#include "../../drivers/net/gianfar.h"

#define PACKET_DEFAULT_RECYCLE_MAX 512

#define PACKET_DATA_HDRLEN	\
	(TPACKET_ALIGN(sizeof(struct pcb_s)) \
		+ sizeof(struct sockaddr_ll))
#define PACKET_FRAME_BMASK		(4096/32)

static void tpum_destruct_skb_err(struct sk_buff *skb);
static void tpum_destruct_skb(struct sk_buff *skb);

/* request callback napi */
struct packet_skb_handler g_psh;
struct packet_sock *po_tx;
int tpum_tx_napi_wt = PACKET_UM_TX_POLL_WT;
int tpum_rx_poll_wt = PACKET_UM_RX_POLL_WT;
int debug_pmal;
unsigned long dbg_pum_sendmsg;
unsigned long dbg_pum_sendmsg_err;
unsigned long pkt_alloc_err;

unsigned long pum_rx_skb_err;
unsigned long pum_rx_skb;
unsigned long pum_tx_skb;
unsigned long pum_tx_skb_err;
unsigned long pmal_tx_kfree;
unsigned long pmal_rx_kfree;
unsigned long pmal_rx_free;
unsigned long pmal_tx_out;


struct packet_filter_node *filter_head;
struct packet_filter_node *filter_tail;

struct packet_asf_cb g_packet_asf;
unsigned int packet_sock_rx;
unsigned int packet_sock_tx;
struct packet_ring_buffer par_ring_rx;
struct packet_ring_buffer par_ring_tx;

#define PACKET_SKB_CB(__skb)	((struct packet_skb_cb *)((__skb)->cb))

#define isprint(c) ((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z') || \
		(c >= '0' && c <= '9'))
static inline void hexdump(const unsigned char *buf, unsigned short len)
{
	char str[80], octet[10];
	int ofs, i, l;

	for (ofs = 0; ofs < len; ofs += 16) {
		sprintf(str, "%03x ", ofs);

		for (i = 0; i < 16; i++) {
			if ((i + ofs) < len)
				sprintf(octet, "%02x ", buf[ofs + i]);
			else
				strcpy(octet, "   ");

			strcat(str, octet);
		}
		strcat(str, "  ");
		l = strlen(str);

		for (i = 0; (i < 16) && ((i + ofs) < len); i++)
			str[l++] = isprint(buf[ofs + i]) ? buf[ofs + i] : '.';

		str[l] = '\0';
		printk(KERN_INFO "%s\n", str);
	}
}

int get_tpkt_sock(struct sock *sk)
{
	int i;
	for (i = 0; i < packet_sock_rx; i++)
		if (tp_sock_data[i].sk == sk)
			return i;
	return -1;
}


int packet_add_buffer_to_list(struct packet_ring_buffer *rb,
					struct pcb_s *ph)
{
	/* Adding the buffer to the list of available buffers */
	rb->list_head->priv.prev_free = (struct packet_buf_s *)ph;
	ph->priv.next_free = (struct packet_buf_s *)rb->list_head;
	ph->priv.prev_free = NULL;
	rb->list_head = ph;
	return 0;
}

struct pcb_s *packet_get_buffer_from_list(
				struct packet_ring_buffer *rb)
{
	struct pcb_s *ph = NULL;

	/* If no free buffer is available, return NULL */
	if (rb->list_tail->priv.prev_free == NULL) {
		pkt_alloc_err++;
		return NULL;
	}
	/* Allocate a buffer from the free list */
	ph = rb->list_tail;
	rb->list_tail = (struct pcb_s *)rb->list_tail->priv.prev_free;
	rb->list_tail->priv.next_free = NULL;
	return ph;
}

static struct pcb_s *packet_alloc_buffer(
	struct packet_ring_buffer *rb)
{
	struct pcb_s *pkt_hdr = NULL;

	pkt_hdr = packet_get_buffer_from_list(rb);
	if (likely(pkt_hdr)) {
		pkt_hdr->pmal_ext.data = pkt_hdr->pmal_ext.head +
			PACKET_UM_HEAD_RESERVE + PACKET_TX_IPSEC_RESERVE;
		pkt_hdr->pmal_ext.tail = pkt_hdr->pmal_ext.data;
		pkt_hdr->pmal_ext.buf_len = 0;
		pkt_hdr->pmal_ext.total_len = 0;
	}
	return pkt_hdr;
}

int packet_free_buffer(struct packet_ring_buffer *rb, unsigned int frame_id)
{
	struct pcb_s *tmp_pmal_buf =
		(struct pcb_s *)rb->frame_addr[frame_id];
	int nr_frags = tmp_pmal_buf->pmal_ext.num_frags;
	tmp_pmal_buf->pmal_ext.num_frags = 0;

	packet_add_buffer_to_list(rb, tmp_pmal_buf);
	while (nr_frags) {
		tmp_pmal_buf =
		(struct pcb_s *)rb->frame_addr[tmp_pmal_buf->priv.next_fid];
		packet_add_buffer_to_list(rb, tmp_pmal_buf);
		nr_frags--;
	}

	return 0;
}

int packet_free_buffer_list(
	struct packet_ring_buffer *rb,
	unsigned int *ph,
	unsigned int nr_buffers)
{
	unsigned int i = 0;
	unsigned int frame_id;

	/* Free 'nr_buffers' buffers to the free list */
	PACKET_RB_LOCK(&rb->frame_pool_lock);
	for (i = 0; i < nr_buffers; i++) {
		frame_id = ph[i];
		packet_free_buffer(rb, frame_id);
		pmal_rx_free++;
	}
	PACKET_RB_UNLOCK(&rb->frame_pool_lock);
	return 0;
}
EXPORT_SYMBOL(packet_free_buffer_list);

unsigned int packet_alloc_buffer_list(
	struct packet_ring_buffer *rb,
	unsigned int ph[],
	unsigned int nr_buffers)
{
	int i;
	struct pcb_s *pkt_hdr = NULL;

	/* Allocate 'nr_buffers' buffers from the free list */
	PACKET_RB_LOCK(&rb->frame_pool_lock);
	for (i = 0; i < nr_buffers; i++) {
		pkt_hdr = packet_alloc_buffer(rb);
		if (unlikely(pkt_hdr == NULL))
			break;
		ph[i] = pkt_hdr->priv.frame_id;
	}
	PACKET_RB_UNLOCK(&rb->frame_pool_lock);
	return i;
}
EXPORT_SYMBOL(packet_alloc_buffer_list);

void *pum_lookup_frame(struct packet_sock *po,
		struct packet_ring_buffer *rb,
		unsigned int position,
		int status)
{
	union {
		struct tpacket_hdr *h1;
		struct tpacket2_hdr *h2;
		void *raw;
	} h;

	/* In case of Packet UM, the look up is done on
	 * Packet UM Ring which does not have the data
	 * buffer but just the tpacket hdr */
	h.raw = (void *)rb->tpacket_hdr_addr[position];
	if (status != __packet_get_status(po, h.raw))
		return NULL;

	return h.raw;
}

void packet_init_frame_desc(struct packet_sock *po,
		struct packet_ring_buffer *rb)
{
	unsigned int pg_vec_pos, frame_offset;
	unsigned int i = 0;
	struct pcb_s *data_hdr = NULL;

	/* Allocate an array of pointers with addresses
	 * to the frame buffers */
	rb->frame_addr = kzalloc((rb->frame_max + 1) * sizeof(unsigned int *),
				GFP_KERNEL);

	PACKET_RB_LOCK_INIT(&rb->frame_pool_lock);
	for (i = 0; i < (rb->frame_max + 1); i++) {
		pg_vec_pos = i / rb->frames_per_block;
		frame_offset = i % rb->frames_per_block;
		rb->frame_addr[i] = (unsigned int)rb->pg_vec[pg_vec_pos]
			+ (frame_offset * rb->frame_size);
		data_hdr = (struct pcb_s *)rb->frame_addr[i];
		data_hdr->priv.frame_id = i;
		data_hdr->priv.virt_addr = rb->frame_addr[i];
		data_hdr->priv.phys_addr =
			virt_to_phys((void *)data_hdr->priv.virt_addr);
		data_hdr->pmal_ext.num_frags = 0;
		data_hdr->pmal_ext.head = PACKET_DATA_HDRLEN
			+ po->tp_reserve;
		data_hdr->pmal_ext.data = data_hdr->pmal_ext.head +
			PACKET_UM_HEAD_RESERVE + PACKET_TX_IPSEC_RESERVE;
		data_hdr->pmal_ext.end =
			rb->frame_size - sizeof(struct skb_shared_info);
		data_hdr->pmal_ext.tail = data_hdr->pmal_ext.data;
		data_hdr->pmal_ext.buf_len = 0;
		flush_dcache_page(virt_to_page(&data_hdr->priv.frame_id));
	}
}


void packet_init_frame_addr(struct packet_sock *po,
		struct packet_ring_buffer *rb)
{
	unsigned int pg_vec_pos, frame_offset;
	unsigned int i = 0;

	/* Allocate an array of pointers with addresses
	 * to the frame buffers */
	rb->frame_addr = kzalloc((rb->frame_max + 1) * sizeof(unsigned int *),
				GFP_KERNEL);

	PACKET_RB_LOCK_INIT(&rb->frame_pool_lock);
	for (i = 0; i < (rb->frame_max + 1); i++) {
		pg_vec_pos = i / rb->frames_per_block;
		frame_offset = i % rb->frames_per_block;
		rb->frame_addr[i] = (unsigned int)rb->pg_vec[pg_vec_pos]
			+ (frame_offset * rb->frame_size);
	}
}

void packet_init_tpacket_hdr_addr(struct packet_sock *po,
		struct packet_ring_buffer *rb)
{
	unsigned int pg_vec_pos, frame_offset;
	unsigned int i = 0;

	/* Allocate an array of pointers with addresses
	 * to the toacket_hdr buffers */
	rb->tpacket_hdr_addr = kzalloc(
		(rb->tpacket_hdr_max + 1) * sizeof(unsigned int *),
		GFP_KERNEL);

	for (i = 0; i < (rb->tpacket_hdr_max + 1); i++) {
		pg_vec_pos = (i / rb->tpacket_hdr_per_block);
		frame_offset = i % rb->tpacket_hdr_per_block;
		rb->tpacket_hdr_addr[i] =
			(unsigned int)rb->pg_vec_tpkt[pg_vec_pos]
			+ (frame_offset * sizeof(struct tpacket2_hdr));
		__packet_set_status(po, (void *)rb->tpacket_hdr_addr[i],
			TP_STATUS_AVAILABLE);
	}
}

static inline void *packet_current_frame(struct packet_sock *po,
		struct packet_ring_buffer *rb,
		int status)
{
	return pum_lookup_frame(po, rb, rb->head, status);
}


static inline void packet_increment_head(struct packet_ring_buffer *buff)
{
	buff->head = buff->head != buff->tpacket_hdr_max ? buff->head+1 : 0;
}

void packet_reset_recycle_queue(void)
{
	struct packet_skb_handler *ph = &g_psh;
	raw_spin_lock_init(&ph->lock);
	ph->recycle_max = PACKET_DEFAULT_RECYCLE_MAX;
	ph->recycle_count = 0;
	ph->recycle_queue = NULL;
}

/*
 * function: packet_free_recycle_queue
 * Reset SKB handler struction and free existance socket buffer
 * and data buffer in the recycling queue.
 */
void packet_free_recycle_queue(int lock_flag)
{
	unsigned long flags = 0;
	struct packet_skb_handler *ph = &g_psh;
	struct sk_buff *clist = NULL;
	struct sk_buff *skb;

	/* Get recycling queue */
	/* just for making sure there is recycle_queue */
	if (lock_flag)
		raw_spin_lock_irqsave(&ph->lock, flags);
	if (ph->recycle_queue) {
		/* pick one from head; most recent one */
		clist = ph->recycle_queue;
		ph->recycle_count = 0;
		ph->recycle_queue = NULL;
	}
	if (lock_flag)
		raw_spin_unlock_irqrestore(&ph->lock, flags);
	while (clist) {
		skb = clist;
		clist = clist->next;
		kfree_skbmem(skb);
	}
}

static inline int packet_recycle_skb(struct sk_buff *skb)
{
	struct packet_skb_handler *ph = &g_psh;

	if (ph->recycle_count >= ph->recycle_max) {
		kfree_skbmem(skb);
		return -1;
	}

	ph->recycle_count++;
	skb->next = NULL;
	if (ph->recycle_queue == NULL) {
		ph->recycle_queue = skb;
		skb->prev = NULL;
		return 0;
	}

	ph->recycle_queue->next = skb;
	skb->prev = ph->recycle_queue;
	ph->recycle_queue = skb;

	return 0;
}

static inline struct sk_buff *packet_reclaim_skb(void)
{
	struct packet_skb_handler *ph = &g_psh;
	struct sk_buff *skb;

	skb = ph->recycle_queue;

	if (skb == NULL)
		return NULL;

	ph->recycle_queue = skb->prev;
	ph->recycle_count--;
	if (ph->recycle_queue != NULL)
		ph->recycle_queue->next = NULL;

	return skb;
}

static inline struct sk_buff *pkt_alloc_skb(struct net_device *dev,
				u8 *data,
				int packet_offset,
				int data_size,
				int header_len)
{
	struct skb_shared_info *shinfo;
	struct sk_buff *skb;
	int node = -1;

	if ((dev) && (dev->dev.parent))
		node = dev_to_node(dev->dev.parent);

	/* Get the HEAD */
	skb = packet_reclaim_skb();
	if (skb == NULL) {
		skb = __alloc_skb_node(GFP_ATOMIC & ~__GFP_DMA, node);
		if (skb == NULL) {
			printk(KERN_INFO"\n%s: Unable to alloc Skb", __func__);
			return NULL;
		}
	}
	/*
	 * Only clear those fields we need to clear, not those that we will
	 * actually initialise below. Hence, don't put any more fields after
	 * the tail pointer in struct sk_buff!
	 */
	memset(skb, 0, offsetof(struct sk_buff, tail));
	skb->truesize = data_size + packet_offset + sizeof(struct sk_buff);
	atomic_set(&skb->users, 1);

	skb->pkt_header = data;
	skb->head = data + packet_offset;

	skb->data = skb->head + header_len;
	skb_reset_tail_pointer(skb);
	skb->end = skb->head + data_size;

	skb->mapped = PF_PACKET_SKB;
	kmemcheck_annotate_bitfield(skb, flags1);
	kmemcheck_annotate_bitfield(skb, flags2);
#ifdef NET_SKBUFF_DATA_USES_OFFSET
	skb->mac_header = ~0U;
#endif

	/* make sure we initialize shinfo sequentially */
	shinfo = skb_shinfo(skb);
	atomic_set(&shinfo->dataref, 1);
	shinfo->nr_frags  = 0;
	shinfo->gso_size = 0;
	shinfo->gso_segs = 0;
	shinfo->gso_type = 0;
	shinfo->ip6_frag_id = 0;
	shinfo->tx_flags.flags = 0;
	skb_frag_list_init(skb);
	memset(&shinfo->hwtstamps, 0, sizeof(shinfo->hwtstamps));

	return skb;
}

struct sk_buff *packet_new_skb(struct net_device *dev)
{
	struct skb_shared_info *shinfo;
	struct sk_buff *skb;
	unsigned int alignamount;
	struct pcb_s *data_hdr = NULL;

	PACKET_RB_LOCK(&par_ring_rx.frame_pool_lock);
	data_hdr = packet_alloc_buffer(&par_ring_rx);

	if (data_hdr == NULL) {
		pum_rx_skb_err++;
		PACKET_RB_UNLOCK(&par_ring_rx.frame_pool_lock);
		return NULL;
	}
	pum_rx_skb++;
	PACKET_RB_UNLOCK(&par_ring_rx.frame_pool_lock);

	skb = pkt_alloc_skb(dev, (u8 *)data_hdr, data_hdr->pmal_ext.head,
			data_hdr->pmal_ext.end - data_hdr->pmal_ext.head,
			PACKET_UM_HEAD_RESERVE);

	if (skb == NULL)
		return NULL;
	alignamount = (((unsigned long) skb->data) % 64);
	if (alignamount)
		skb_reserve(skb, 64 - alignamount);

	shinfo = skb_shinfo(skb);
	shinfo->destructor_arg = skb->pkt_header;
	skb->destructor = tpum_destruct_skb_err;

	return skb;
}
EXPORT_SYMBOL(packet_new_skb);


void packet_kfree_skb(struct sk_buff *skb)
{
	if (unlikely(!skb))
		return;

	if (skb->mapped != PF_PACKET_SKB) {
		kfree_skb(skb);
		return;
	}

	skb_release_head_state(skb);
	skb->destructor = 0;
	packet_recycle_skb(skb);
}
EXPORT_SYMBOL(packet_kfree_skb);

void pmal_convert_to_sg(struct sk_buff *skb)
{
	struct pcb_s *pmal_buff_hdr =
		(struct pcb_s *)(skb->pkt_header);
	struct pcb_s *tmp1_pmal_hdr = NULL;
	struct pcb_s *tmp2_pmal_hdr = NULL;
	int i = 0;
	int nr_frags = 0;
	struct skb_shared_info *sh_info = skb_shinfo(skb);
	unsigned long data_off = 0;
	unsigned int alignamount = 0;
	struct sk_buff *tmp_skb = skb;

	pmal_buff_hdr->pmal_ext.total_len = skb->len;
	pmal_buff_hdr->pmal_ext.num_frags = 0;
	pmal_buff_hdr->priv.next_fid = 0;
	pmal_buff_hdr->pmal_ext.buf_len = skb->len - skb->data_len;
	pmal_buff_hdr->pmal_ext.tail = skb->tail
			- (unsigned char *)pmal_buff_hdr;

	if ((sh_info->nr_frags == 0) &&
			(sh_info->frag_list == NULL))
		return ;

	data_off = PACKET_DATA_HDRLEN + PACKET_UM_HEAD_RESERVE;
	alignamount = data_off % 64;
	if (alignamount)
		data_off += (64 - alignamount);
	tmp1_pmal_hdr = pmal_buff_hdr;

	if (sh_info->frag_list == NULL)
		goto __nr_frags;

	tmp_skb->next = sh_info->frag_list;

	do {
		tmp1_pmal_hdr->priv.first_fid = pmal_buff_hdr->priv.frame_id;
		tmp1_pmal_hdr->pmal_ext.data = tmp_skb->data
			- (unsigned char *)tmp1_pmal_hdr;
		tmp1_pmal_hdr->pmal_ext.tail = tmp_skb->tail
			- (unsigned char *)tmp1_pmal_hdr;
		tmp1_pmal_hdr->pmal_ext.buf_len =
			tmp_skb->len - tmp_skb->data_len;

		if (tmp_skb->next)
			tmp1_pmal_hdr =
				(struct pcb_s *)tmp_skb->next->pkt_header;
		else
			tmp1_pmal_hdr = NULL;
		tmp_skb = tmp_skb->next;
		pmal_buff_hdr->pmal_ext.num_frags++;
	} while (tmp_skb);

	pmal_buff_hdr->pmal_ext.num_frags--;
	return;

__nr_frags:
	/* For each skb in the fraglist */
#ifdef CONFIG_GIANFAR_SG_SUPPORT
	tmp_skb->next =  skb_shinfo(skb)->gianfar_destructor_arg;
#endif
	tmp1_pmal_hdr->priv.first_fid = pmal_buff_hdr->priv.frame_id;
	tmp1_pmal_hdr->pmal_ext.data = skb->data
		- (unsigned char *)tmp1_pmal_hdr;
	tmp1_pmal_hdr->pmal_ext.tail = skb->tail
		- (unsigned char *)tmp1_pmal_hdr;
	tmp1_pmal_hdr->pmal_ext.buf_len = skb->len - skb->data_len;
	tmp1_pmal_hdr->pmal_ext.total_len = tmp1_pmal_hdr->pmal_ext.buf_len;

	nr_frags = sh_info->nr_frags;


	/* For nr_frags in this skb */
	for (i = 0; i < nr_frags; i++) {
		/* Since there is list of skb which together constituted this
		 * packet so we shall use the pkt_header in each of these skb
		 * to get the pmal_hdr as one cannot take out the pmal header
		 * from the data pointer in the data buffer which we get from
		 * frags array as the data pointer is going to point to diff
		 * offset for diff scenarios */
		tmp2_pmal_hdr = (struct pcb_s *)tmp_skb->next->pkt_header;
		tmp2_pmal_hdr->pmal_ext.data =
			(unsigned char *)page_address(sh_info->frags[i].page)
			+ sh_info->frags[i].page_offset
			- (unsigned char *)tmp2_pmal_hdr;
		tmp2_pmal_hdr->priv.first_fid = pmal_buff_hdr->priv.frame_id;
		tmp2_pmal_hdr->pmal_ext.buf_len = sh_info->frags[i].size;
		tmp2_pmal_hdr->pmal_ext.tail = tmp2_pmal_hdr->pmal_ext.data
					+ tmp2_pmal_hdr->pmal_ext.buf_len;
		pmal_buff_hdr->pmal_ext.num_frags++;
		pmal_buff_hdr->pmal_ext.total_len +=
			tmp2_pmal_hdr->pmal_ext.buf_len;

		tmp1_pmal_hdr->priv.next_fid = tmp2_pmal_hdr->priv.frame_id;
		tmp1_pmal_hdr = tmp2_pmal_hdr;
		tmp_skb = tmp_skb->next;
	}
	return ;
}

void pmal_convert_from_sg(struct net_device *dev,
		struct sock *sk,
		struct sk_buff *skb)
{
	struct pcb_s *pmal_buff_hdr =
		(struct pcb_s *)(skb->pkt_header);
	struct skb_shared_info *sh_info = NULL;
	int num_frags = pmal_buff_hdr->pmal_ext.num_frags;
	unsigned int next_fid = pmal_buff_hdr->priv.next_fid;
	unsigned char *offset = NULL;
	int i = 0;
	int k = 0;

	if (!num_frags || (num_frags > MAX_SKB_FRAGS))
		return;
	sh_info = skb_shinfo(skb);
	do {
		pmal_buff_hdr =
			(struct pcb_s *)par_ring_tx.frame_addr[next_fid];
		sh_info->frags[i].page =
			virt_to_page(pmal_buff_hdr);
		offset = (unsigned char *)pmal_buff_hdr +
			pmal_buff_hdr->pmal_ext.data;
		sh_info->frags[i].page_offset =
			offset_in_page(offset);
		sh_info->frags[i].size =
			pmal_buff_hdr->pmal_ext.buf_len;
		skb->data_len += pmal_buff_hdr->pmal_ext.buf_len;
		sh_info->nr_frags++;
		next_fid = pmal_buff_hdr->priv.next_fid;
		i++;
		k++;
	} while (k != num_frags);

	return;
}

int tpum_rcv(struct sk_buff *skb, struct net_device *dev,
		struct packet_type *pt, struct net_device *orig_dev)
{
	struct sock *sk;
	struct packet_sock *po;
	struct sockaddr_ll *sll;
	union {
		struct tpacket_hdr *h1;
		struct tpacket2_hdr *h2;
		void *raw;
	} h;
	u8 *skb_head = skb->data;
	int skb_len = skb->len;
	unsigned int snaplen;
	unsigned long status = TP_STATUS_LOSING|TP_STATUS_USER;
	static unsigned long pkt_rcv_cnt;
	unsigned short macoff, netoff, hdrlen;
	struct sk_buff *copy_skb = NULL;
	struct pcb_s *pkt_data_hdr;
	int tpkt_sock_index;

	if (skb->pkt_type == PACKET_LOOPBACK)
		goto drop;

	/* Drop the packet if its not meant for PMAL */
	if (skb->mapped == 0)
		goto drop;

	/* extract the sock info from the packet_type */
	sk = pt->af_packet_priv;
	po = pkt_sk(sk);

	if (!net_eq(dev_net(dev), sock_net(sk)))
		goto drop;

	if (dev->header_ops) {
		if (sk->sk_type != SOCK_DGRAM)
			skb_push(skb, skb->data - skb_mac_header(skb));
		else if (skb->pkt_type == PACKET_OUTGOING) {
			/* Special case: outgoing packets
			 * have ll header at head */
			skb_pull(skb, skb_network_offset(skb));
		}
	}

	snaplen = skb->len;

	if (sk->sk_type == SOCK_DGRAM) {
		netoff = (unsigned short)(skb->data - skb->pkt_header);
		macoff = netoff - skb->mac_len;
	} else {
		macoff = (unsigned short)(skb->data - skb->pkt_header);
		netoff = macoff + skb->mac_len;
		skb_pull(skb, skb->mac_len);
	}

	spin_lock(&sk->sk_receive_queue.lock);

	/* Find the index in the Packet UM RIng where we can
	 * enqueue the Frame buffer */
	h.raw = packet_current_frame(po, &po->rx_ring, TP_STATUS_KERNEL);
	if (!h.raw)
		goto ring_is_full;


	/* We just need to store the frame id instead of the frame pointer in
	 * the Ring buffer */
	h.h2->tp_frame_id =
			((struct pcb_s *)(skb->pkt_header))->priv.frame_id;
	flush_dcache_page(virt_to_page(&h.h2->tp_frame_id));

	/* Increment the inpkt stats */
	packet_increment_head(&po->rx_ring);
	po->stats.tp_packets++;
	if (!po->stats.tp_drops)
		status &= ~TP_STATUS_LOSING;
	pkt_data_hdr = (struct pcb_s *)skb->pkt_header;
	if (!pkt_data_hdr) {
		printk(KERN_INFO"\n %s ERROR NULL pkt_header", __func__);
		goto drop;
	}

	pkt_data_hdr->pmal_ext.data = netoff;
	pmal_convert_to_sg(skb);

	spin_unlock(&sk->sk_receive_queue.lock);

	h.h2->tp_len = skb->len;
	h.h2->tp_snaplen = snaplen;
	h.h2->tp_mac = macoff;
	h.h2->tp_net = netoff;
	if (vlan_tx_tag_present(skb))
		h.h2->tp_vlan_tci = vlan_tx_tag_get(skb);
	else
		h.h2->tp_vlan_tci = 0;

	hdrlen = (unsigned short)sizeof(struct pcb_s);
	sll = (struct sockaddr_ll *)((char *)pkt_data_hdr +
						TPACKET_ALIGN(hdrlen));
	sll->sll_halen = (unsigned char)dev_parse_header(skb, sll->sll_addr);
	sll->sll_family = AF_PACKET;
	sll->sll_hatype = dev->type;
	sll->sll_protocol = skb->protocol;
	sll->sll_pkttype = skb->pkt_type;
	if (unlikely(po->origdev))
		sll->sll_ifindex = orig_dev->ifindex;
	else
		sll->sll_ifindex = dev->ifindex;

	/* Change the status of the Ring Buffer to TP_STATUS_USER
	 * so that when the user polls on this Ring, he knows that
	 * a frame has been received */
	__packet_set_status(po, h.raw, status);

	tpkt_sock_index = get_tpkt_sock(sk);
	if (tpkt_sock_index < 0)
		goto drop;
	pkt_rcv_cnt++;
	if (!(pkt_rcv_cnt % tpum_rx_poll_wt)) {
		sk->sk_data_ready(sk, 0);
		tp_sock_data[tpkt_sock_index].data_ready = 0;
	} else
		tp_sock_data[tpkt_sock_index].data_ready = 1;

	if (!timer_pending(&tpkt_timer))
		mod_timer(&tpkt_timer, tpkt_timer.expires);

	skb->destructor = 0;
drop_n_restore:
	if (skb_head != skb->data && skb_shared(skb)) {
		skb->data = skb_head;
		skb->len = skb_len;
	}
drop:
	packet_kfree_skb(skb);
	return 0;

ring_is_full:
	po->stats.tp_drops++;
	spin_unlock(&sk->sk_receive_queue.lock);

	sk->sk_data_ready(sk, 0);
	kfree_skb(copy_skb);
	goto drop_n_restore;
}

static void tpum_destruct_skb(struct sk_buff *skb)
{
	struct packet_sock *po = pkt_sk(skb->sk);
	struct tpacket2_hdr *ph;
	struct pcb_s *pcb;

	BUG_ON(skb == NULL);
	if (unlikely(!po)) {
		printk(KERN_WARNING"\n%s - null packet_sock", __func__);
		return;
	}
	if (!skb->mapped)
		printk(KERN_WARNING"\n%s - skb unmapped", __func__);

	if (likely(po->tx_ring.pg_vec)) {
		ph = skb_shinfo(skb)->destructor_arg;
		if (ph == NULL) {
			printk("\n%s - ERROR-dest_arg=NULL (skb 0x%x-mapped=%d)",
				__func__, (unsigned int)skb, skb->mapped);
			goto temp1;
		}
#if 0
		BUG_ON(__packet_get_status(po, ph) != TP_STATUS_SENDING);
		BUG_ON(atomic_read(&po->tx_ring.pending) == 0);
#endif
		atomic_dec(&po->tx_ring.pending);

		pcb = (struct pcb_s *)skb->pkt_header;
		if (pcb == NULL) {
			printk("\n%s - ERROR-skb_pkt=NULL (skb 0x%x-mapped=%d) dest_arg=0x%x ",
				__func__, (unsigned int)skb, skb->mapped,
				(unsigned int)skb_shinfo(skb)->destructor_arg);
			goto temp1;
		}

		if ((skb_shinfo(skb)->nr_frags != 0) &&
				 (pcb->pmal_ext.num_frags == 0))
			pmal_convert_to_sg(skb);

		PACKET_RB_LOCK(&par_ring_tx.frame_pool_lock);
		pmal_tx_kfree++;
		packet_free_buffer(&par_ring_tx, ph->tp_frame_id);
		PACKET_RB_UNLOCK(&par_ring_tx.frame_pool_lock);

		/* Change the Status of the Ring BUffer in TX
		 * Packet UM Ring to Available since the Frame
		 * has already been transmitted */
		ph->tp_frame_id = 0xffff;
		__packet_set_status(po, ph, TP_STATUS_AVAILABLE);
	}
temp1:
	skb->destructor = 0;
}

static void tpum_destruct_skb_err(struct sk_buff *skb)
{
	struct packet_sock *po = NULL;
	struct pcb_s *pcb;
	BUG_ON(skb == NULL);

	pcb = (struct pcb_s *)skb->pkt_header;
	if (pcb == NULL) {
		printk("\n%s - ERROR-skb_pkt=NULL (skb 0x%x-mapped=%d) dest_arg=0x%x ",
			__func__, (unsigned int)skb, skb->mapped,
			(unsigned int)skb_shinfo(skb)->destructor_arg);

		goto temp1;
	}

	if ((skb_shinfo(skb)->nr_frags != 0) && (pcb->pmal_ext.num_frags == 0))
			pmal_convert_to_sg(skb);

	PACKET_RB_LOCK(&par_ring_rx.frame_pool_lock);
	pmal_rx_kfree++;
	packet_free_buffer(&par_ring_rx, pcb->priv.frame_id);
	PACKET_RB_UNLOCK(&par_ring_rx.frame_pool_lock);

	po = pkt_sk(skb->sk);
	if (po)
		po->stats.tp_drops++;

temp1:
	skb->mapped = 0;
	skb->destructor = 0;
	skb->pkt_header = 0;
}

int tpum_snd_napi(struct napi_struct *napi, int budget)
{
	tpum_snd((struct packet_sock *)napi->arg0,
			&((struct packet_sock *)napi->arg0)->msg);
	napi_complete(napi);
	return 0;
}

int tpum_snd(struct packet_sock *po, struct msghdr *msg)
{
	struct socket *sock;
	struct sock *sk;
	struct sk_buff *skb;
	struct net_device *dev;
	__be16 proto;
	int ifindex, err, reserve = 0;
	void *frame;
	struct sockaddr_ll *saddr = (struct sockaddr_ll *)msg->msg_name;
	int tp_len, size_max;
	unsigned char *addr = NULL;
	int len_sum = 0;
	int tx_poll_wt = 0;
	union {
		struct tpacket_hdr *h1;
		struct tpacket2_hdr *h2;
		void *raw;
	} ph;
	struct pcb_s *data_hdr;
	struct packet_ring_buffer *rb = &po->tx_ring;


	sock = po->sk.sk_socket;
	sk = &po->sk;

	mutex_lock(&po->pg_vec_lock);

	if (saddr == NULL) {
		ifindex	= po->ifindex;
		proto	= po->num;

	} else {
		err = -EINVAL;
		if (msg->msg_namelen < sizeof(struct sockaddr_ll))
			goto out;
		if (msg->msg_namelen < (saddr->sll_halen
					+ offsetof(struct sockaddr_ll,
						sll_addr)))
			goto out;
		ifindex	= saddr->sll_ifindex;
		proto	= saddr->sll_protocol;
		addr	= saddr->sll_addr;
	}

	/* Find the eth interface on which to send the packet */
	dev = dev_get_by_index(sock_net(&po->sk), ifindex);
	err = -ENXIO;
	if (unlikely(dev == NULL))
		goto out;

	reserve = dev->hard_header_len;

	err = -ENETDOWN;
	if (unlikely(!(dev->flags & IFF_UP)))
		goto out_put;

	size_max = po->tx_ring.frame_size
		- (po->tp_hdrlen - sizeof(struct sockaddr_ll));

	if (size_max > dev->mtu + reserve)
		size_max = dev->mtu + reserve;

	do {
		frame = packet_current_frame(po, &po->tx_ring,
				TP_STATUS_SEND_REQUEST);

		/* No more packets in the TX Packet UM Ring
		 * is there which needs to be rtansmitted */
		if (unlikely(frame == NULL))
			continue;

		ph.raw = frame;
		data_hdr = (struct pcb_s *)
				rb->frame_addr[ph.h2->tp_frame_id];
		tp_len = ph.h2->tp_len;

		skb = pkt_alloc_skb(dev, (u8 *)data_hdr,
			data_hdr->pmal_ext.head,
			data_hdr->pmal_ext.end - data_hdr->pmal_ext.head,
			data_hdr->pmal_ext.data - data_hdr->pmal_ext.head);

		if (unlikely(skb == NULL)) {
			printk(KERN_WARNING "skb is not available\n");
			po->stats.tp_drops++;
			packet_free_buffer(&po->tx_ring, ph.h2->tp_frame_id);
			goto out_put;
		}
		skb_set_owner_w(skb, sk);
		skb->tail = skb->data + data_hdr->pmal_ext.buf_len;

		skb->mapped = PF_PACKET_SKB;
		skb_shinfo(skb)->destructor_arg = frame;
		skb->destructor = tpum_destruct_skb;

		/* Point skb->data to ph */
		skb->protocol = proto;
		skb->dev = dev;
		skb->priority = po->sk.sk_priority;
		skb_set_network_header(skb, 0);
		skb->len = tp_len;
		if (sock->type == SOCK_DGRAM) {
			err = dev_hard_header(skb, dev, ntohs(proto), addr,
					NULL, skb->len);
			if (unlikely(err < 0)) {
				printk(KERN_WARNING "\ndev_hard_header failed");
				goto out_xmit;
			}
		} else {
			skb_push(skb, ETH_HLEN);
		}

		skb_reset_mac_header(skb);
		skb->mac_len = (__u16)skb_network_offset(skb);

		if (((struct pcb_s *)(skb->pkt_header))->pmal_ext.num_frags >
				PACKET_UM_MAX_SG_FRAGS)
			goto out_xmit;
		pmal_convert_from_sg(dev, sk, skb);

		__packet_set_status(po, frame, TP_STATUS_SENDING);
		atomic_inc(&po->tx_ring.pending);
		po->stats.tp_packets++;

		/* Call ASF Tx function */
		if (g_packet_asf.packet_asf_tx_cb)
			g_packet_asf.packet_asf_tx_cb(skb, NULL);
		else {
			skb_set_queue_mapping(skb, 0);

			err = dev->netdev_ops->ndo_start_xmit(skb, dev);
			if (unlikely(err > 0)) {
				err = net_xmit_errno(err);
				if (err != 0)
					goto out_xmit;
			}
		}

		/* Increament the Head of the Packet UM Ring */
		packet_increment_head(&po->tx_ring);
		len_sum += tp_len;
		tx_poll_wt++;
	} while (likely((frame != NULL) &&
				(tx_poll_wt < tpum_tx_napi_wt)));
	err = tx_poll_wt;
	pmal_tx_out++;
	goto out_put;

out_xmit:
	printk(KERN_ERR"%s Need to drop the SKB, err %x", __func__, err);
	po->stats.tp_drops++;
	packet_kfree_skb(skb);
out_put:
	dev_put(dev);
out:
	mutex_unlock(&po->pg_vec_lock);
	return err;
}


int packet_asft_flush_connection(void *ctxt)
{
	if (g_packet_asf.packet_asf_config)
		return g_packet_asf.packet_asf_config(ASF_PACKET_FLUSH_CONN,
				ctxt);
	return 0;
}

int packet_asft_add_connection(struct packet_con_s *conn)
{
	if (g_packet_asf.packet_asf_config)
		return g_packet_asf.packet_asf_config(ASF_PACKET_CONN_ADD,
				conn);
	return 0;
}

int packet_asft_del_connection(struct packet_con_s *conn)
{
	if (g_packet_asf.packet_asf_config)
		g_packet_asf.packet_asf_config(ASF_PACKET_CONN_DEL, conn);
	return 0;
}


/*
 *	Pull a packet from our receive queue and hand it to the user.
 *	If necessary we block.
 */

bool packet_is_valid_flow(struct packet_con_s *flow,
			enum packet_flow_type flow_type)
{
	unsigned int i = 0;
	struct packet_filter_node *node;

	node = filter_head;

	while (node) {
		if (node->filter.ip_addr) {
			if (flow_type == e_RX_UDP) {
				if (flow->daddr.sin_addr.s_addr
						!= node->filter.ip_addr)
					goto incr_node;

				if (node->filter.udp_ports[0])
					for (i = 0;
					 i < PACKET_UM_MAX_FILTER_PORTS; i++) {
						if (node->filter.udp_ports[i]
							== flow->daddr.sin_port)
							goto create_conn;
					}
				else
					goto create_conn;

			} else {
				if (flow->saddr.sin_addr.s_addr
						!= node->filter.ip_addr)
					goto incr_node;

				if (node->filter.udp_ports[0])
					for (i = 0;
					i < PACKET_UM_MAX_FILTER_PORTS; i++) {
						if (node->filter.udp_ports[i]
							== flow->saddr.sin_port)
							goto create_conn;
					}
				else
					goto create_conn;

			}
		} else {
			if (flow_type == e_RX_UDP) {
				for (i = 0;
					i < PACKET_UM_MAX_FILTER_PORTS; i++) {
					if (node->filter.udp_ports[i]
							== flow->daddr.sin_port)
						goto create_conn;
				}
			} else {
				for (i = 0;
					i < PACKET_UM_MAX_FILTER_PORTS; i++) {
					if (node->filter.udp_ports[i]
							== flow->saddr.sin_port)
						goto create_conn;
				}
			}
		}
incr_node:
		node = node->next;
	}
	return 0;

create_conn:
	flow->ctxt = node->sock_cnxt;
	return 1;
}
EXPORT_SYMBOL(packet_is_valid_flow);

int pmal_register_tx_hook(struct packet_asf_cb *asfstrct)
{
	if (packet_sock_rx || packet_sock_tx)
		return -EBUSY;
	g_packet_asf.packet_asf_tx_cb = asfstrct->packet_asf_tx_cb;
	g_packet_asf.packet_asf_config = asfstrct->packet_asf_config;
	return 0;
}
EXPORT_SYMBOL(pmal_register_tx_hook);

void pmal_unregister_tx_hook()
{
	g_packet_asf.packet_asf_tx_cb = NULL;
	g_packet_asf.packet_asf_config = NULL;
}
EXPORT_SYMBOL(pmal_unregister_tx_hook);

int pmal_netif_receive_skb(struct sk_buff *skb)
{
	struct sk_buff *nskb = skb;

	if (skb->mapped == PF_PACKET_SKB) {
		/* Allocate new skb from kernel pool */
		nskb = skb_copy(skb, GFP_ATOMIC);
		packet_kfree_skb(skb);
		if (!nskb)
			return NET_RX_DROP;
		nskb->mapped = 0;
	}
	return netif_receive_skb(nskb);
}
EXPORT_SYMBOL(pmal_netif_receive_skb);

int pmal_match_filter(struct sk_buff *skb)
{
	struct iphdr	*iph;
	struct packet_con_s conn;

	skb->network_header = skb->data;
	iph = ip_hdr(skb);

	if (iph->protocol != IPPROTO_UDP)
		return -1;

	conn.daddr.sin_addr.s_addr = iph->daddr;
	conn.daddr.sin_port =
		(*(unsigned short *) (((char *) (iph) + iph->ihl*4) + 2));

	conn.proto = iph->protocol;
	conn.ctxt = 0;

	if (packet_is_valid_flow(&conn, e_RX_UDP)) {
		skb->pmal_ctxt = conn.ctxt;
		return 0;
	}

	return -1;
}

int pmal_receive_skb(struct sk_buff *skb)
{
	struct packet_type *ptype;

	if (skb->mapped != PF_PACKET_SKB) {
		printk(KERN_INFO"\n%s: not pf packet, dropping", __func__);
		show_stack(current, (unsigned long *)__get_SP());
		kfree_skb(skb);
		return NET_RX_DROP;
	}

	skb_reset_mac_header(skb);
	skb_pull(skb, ETH_HLEN);

	if (!skb->pmal_ctxt) {
		if (pmal_match_filter(skb) < 0) {
			printk(KERN_INFO"\n Filter Criteria Failed");
			return pmal_netif_receive_skb(skb);
		}
	}

	ptype = &((struct packet_sock *)(skb->pmal_ctxt))->prot_hook;
	ptype->func(skb, skb->dev, ptype, skb->dev);
	return 0;
}
EXPORT_SYMBOL(pmal_receive_skb);

