/*
 * File:	af_packet_um_sysfs.c
 *
 * Description: SysFS interface for configuring the PACKET_UM options
 *
 * Authors:	Sachin Saxena <sachin.saxena@freescale.com>
 *
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 *	Copyright 2011-2012 Freescale Semiconductor, Inc.
*/

#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/sysctl.h>
#include <net/sock.h>
#include <linux/if_pmal.h>
#include "af_packet.h"

static ssize_t rx_poll_wt_show(struct kobject *kobj,
				struct kobj_attribute *attr, char *buf)
{
	return sprintf(buf, "%d\n", tpum_rx_poll_wt);;
}


static ssize_t rx_poll_wt_store(struct kobject *kobj,
				struct kobj_attribute *attr,
				const char *buf, size_t count)
{
	int val;

	sscanf(buf, "%d", &val);
	if (val <= 0) {
		printk(KERN_INFO "[%d] is INVALID... Please enter"
					" a value > 0 !\n", val);
		return -EINVAL;
	}

	if (val != tpum_rx_poll_wt) {
		printk(KERN_INFO "\nRx Poll Weight[%d] changed to [%d]\n",
						tpum_rx_poll_wt, val);
		tpum_rx_poll_wt = val;
	}

	return count;
}


static ssize_t tx_napi_wt_show(struct kobject *kobj,
				struct kobj_attribute *attr, char *buf)
{
	return sprintf(buf, "%d\n", tpum_tx_napi_wt);;
}


static ssize_t tx_napi_wt_store(struct kobject *kobj,
				struct kobj_attribute *attr,
				const char *buf, size_t count)
{
	int val;

	sscanf(buf, "%d", &val);
	if (val <= 0) {
		printk(KERN_INFO "[%d] is INVALID... Please enter"
					" a value > 0 !\n", val);
		return -EINVAL;
	}

	if (val != tpum_tx_napi_wt) {
		printk(KERN_INFO "\nTx Napi Weight[%d] changed to [%d]\n",
						tpum_tx_napi_wt, val);
		tpum_tx_napi_wt = val;
	}

	return count;
}

static ssize_t debug_pmal_wt_show(struct kobject *kobj,
				struct kobj_attribute *attr, char *buf)
{
	printk(KERN_INFO "\nsend=%lu tx_out=%lu s_err=%lu alloc_err=%lu\n"
			"rx=%lu, rx_err=%lu tx=%lu, tx_err=%lu\n"
			"tx_fr=%lu, rx_fr=%lu rx_fr1=%lu\n",
			dbg_pum_sendmsg, pmal_tx_out, dbg_pum_sendmsg_err, pkt_alloc_err,
			pum_rx_skb, pum_rx_skb_err, pum_tx_skb,
			pum_tx_skb_err, pmal_tx_kfree, pmal_rx_kfree,
			pmal_rx_free);
	return sprintf(buf, "%d\n", debug_pmal);
}

static struct kobj_attribute tx_napi_wt_attr = \
	__ATTR(tx_napi_wt, 0644,
		tx_napi_wt_show, tx_napi_wt_store);

static struct kobj_attribute rx_poll_wt_attr = \
	__ATTR(rx_poll_wt, 0644,
		rx_poll_wt_show, rx_poll_wt_store);

static struct kobj_attribute debug_pmal_attr = \
	__ATTR(debug_pmal, 0444,
		debug_pmal_wt_show, NULL);

static struct attribute *packet_um_attrs[] = {
	&tx_napi_wt_attr.attr,
	&rx_poll_wt_attr.attr,
	&debug_pmal_attr.attr,
	NULL
};

static struct attribute_group packet_um_attr_group = {
	.attrs = packet_um_attrs,
};

struct kobject *packet_um_kobj;
EXPORT_SYMBOL_GPL(packet_um_kobj);



int packet_um_sysfs_init(void)
{
	int error;

	packet_um_kobj = kobject_create_and_add("packet_um", NULL);
	if (!packet_um_kobj) {
		error = -ENOMEM;
		goto exit;
	}

	error = sysfs_create_group(packet_um_kobj, &packet_um_attr_group);
	if (error)
		goto attr_exit;

	return 0;

attr_exit:
	kobject_put(packet_um_kobj);
exit:
	return error;
}

int asfctrl_sysfs_exit(void)
{
	sysfs_remove_group(packet_um_kobj, &packet_um_attr_group);
	kobject_put(packet_um_kobj);
	return 0;
}

